@@ -6,6 +6,8 @@
 import shlex, os, re
 import subprocess
 
+import almost_make.utils.shellUtil.globber as globber
+
 PRECEDENCE_LIST = [ '||', '&&', ";", '|', '>', '2>&1', '&' ]
 TWO_ARGUMENTS = { "||", "&&", ";", '|', '>' }
 PIPE_OUT_PERMISSIONS = 0o660
@@ -16,7 +18,8 @@ SYSTEM_SHELL = "system-shell"
 USE_SYSTEM_PIPE = "use-system-pipe"
 
 class ShellState:
-    cwd = None
+    def __init__(self):
+        self.cwd = os.getcwd()
 
 # Get the number of [char] at the beginning of [text].
 def getStartingCount(text, char):
@@ -82,7 +85,7 @@ def stripQuotes(text):
         return text
     result = []
     firstChar = text[0]
-    lastChar = text[len(text) - 1]
+    lastChar = text[-1]
     
     if firstChar != lastChar:
         return text
@@ -91,6 +94,35 @@ def stripQuotes(text):
         return text[1 : len(text) - 1]
     return text
 
+# Return whether text begins and ends with the same quoting character,
+# **and** the quoting is un-interupted. For example, "a, b, c" -> True,
+# "a, b, c' -> False, "a, b," "c" -> False.
+def isQuoted(text):
+    if len(text) < 2:
+        return False
+    if text[0] != text[-1]: # Must start and end with the same character.
+        return False
+    if not text[0] in { '"', "'" }:
+        return False
+    
+    startQuote = text[0]
+
+    escaped = False
+
+    text = text[1:-1] # Remove starting & ending...
+
+    # If something would have broken us out of the quote...
+    for char in text:
+        if char == startQuote and not escaped:
+            return False # ... then it isn't completely quoted.
+        elif char == '\\' and not escaped:
+            escaped = True
+        elif escaped:
+            escaped = False
+    
+    # If the ending quote would have been escaped, it isn't quoted.
+    return not escaped
+
 # Cluster elements in [splitText] based on precedence.
 # E.g. ['a', '||', 'b', '&&', 'c', '-l'] -> [['a'], '||', [['b'], '&&', ['c', '-l']]].
 def cluster(splitText, level=0):
@@ -151,7 +183,7 @@ def collapse(clustered):
         return result
     elif type(clustered[0]) == str:
         for part in clustered:
-            if SPACE_CHARS.search(part.strip()) != None:
+            if SPACE_CHARS.search(part.strip()) != None and not isQuoted(part):
                 result += " " + shlex.quote(part)
             else:
                 result += " " + part
@@ -161,22 +193,19 @@ def collapse(clustered):
     
     return result.strip()
 
-# Glob a single argument, [arg], from the
-# perspective of [cwd].
-# See: https://www.gnu.org/software/bash/manual/bash.html#Filename-Expansion
-def glob(arg, cwd):
-    inQuote = None
-
-    # To-do!!!
-
 # Glob all arguments in args, excluding the first and 
 # quoted arguments.
 def globArgs(args, state):
     cwd = os.path.abspath(state.cwd or '.')
     result = []
+    isFirst = True
     
     for arg in args:
-        result.append(glob(arg, cwd))
+        if not isQuoted(arg.strip()) and not isFirst:
+            result.extend(globber.glob(arg, cwd))
+        else:
+            result.append(arg)
+            isFirst = False
     
     return result
 
@@ -204,7 +233,10 @@ def rawRun(args, customCommands={}, flags=[], stdin=None, stdout=None, stderr=No
     
     sysShell = SYSTEM_SHELL in flags or None
 
-    # args = globArgs(args, state)
+    if not sysShell:
+        args = globArgs(args, state)
+        args = [ stripQuotes(arg) for arg in args ]
+
     command = args[0].strip()
 
     if command in customCommands:
@@ -305,11 +337,11 @@ def shSplit(text):
             elif inQuote == None:
                 inQuote = char
         elif char in [ '(', ')', '|', '&', '>', ';', ' ', '\t', '\n' ] and inQuote == None:
-            result.append(stripQuotes(buff[:len(buff)-1]))
+            result.append(buff[:-1])
             result.append(char)
             buff = ""
     
-    result.append(stripQuotes(buff))
+    result.append(buff)
     
     buff = ''
     filtered = []
@@ -429,7 +461,8 @@ if __name__ == "__main__":
     assertEql(shSplit("1||2&&3"), ["1", "|", "|", "2", "&", "&", "3"], "More complicated, no-space splitting.")
     assertEql(shSplit("1 || 2"), ["1", "|", "|", "2"], "Pipe splitting with spaces.")
     assertEql(shSplit("ls -la"), ["ls", "-la"], "Space-splitting")
-    assertEql(shSplit("'ls -la'"), ["ls -la"], "Quoted space-(not)splitting")
+    assertEql(shSplit("'ls -la'"), ["'ls -la'"], "Quoted space-(not)splitting")
+    assertEql(shSplit("'ls' '-la'"), ["'ls'", "'-la'"], "Quoted parts splitting")
     assertEql(shSplit("1  \t\n > 2 &\n   & 1>2     "), ["1", ">", "2", "&", "&", "1", ">", "2"], "Lots of spaces!")
     assertEql(shSplit("1;2"), ["1", ";", "2"], "Does it correctly split on semi-colons?")
     assertEql(shSplit("1(2)"), ["1", "(", "2", ")"], "What about parentheses?")
@@ -438,8 +471,8 @@ if __name__ == "__main__":
     assertEql(filterSplitList(shSplit("ls || ls")), ["ls", "||", "ls"], "Shlex replacement test 1!")
     assertEql(filterSplitList(shSplit("ls && ps")), ["ls", "&&", "ps"], "Shlex replacement test 2!")
     assertEql(filterSplitList(shSplit("ls; ps")), ["ls", ";", "ps"], "Shlex replacement test 3!")
-    assertEql(filterSplitList(shSplit("ls; (ps 2>&1 | grep 'foo && not foo')")), ["ls", ";", "(", "ps", "2>&1", "|", "grep", "foo && not foo", ")"], "Shlex replacement test 4!")
-    assertEql(filterSplitList(shSplit("\"ls; (ps 2>&1 | grep 'foo && not foo')\"")), ["ls; (ps 2>&1 | grep 'foo && not foo')"], "Shlex replacement test 5!")
+    assertEql(filterSplitList(shSplit("ls; (ps 2>&1 | grep 'foo && not foo')")), ["ls", ";", "(", "ps", "2>&1", "|", "grep", "'foo && not foo'", ")"], "Shlex replacement test 4!")
+    assertEql(filterSplitList(shSplit("\"ls; (ps 2>&1 | grep 'foo && not foo')\"")), ["\"ls; (ps 2>&1 | grep 'foo && not foo')\""], "Shlex replacement test 5!")
     assertEql(filterSplitList(shSplit("make CFLAGS=thing LDFLAGS= CC=cc GCC=g++ A= TEST=33")), 
         [ "make", "CFLAGS=thing", "LDFLAGS=", "CC=cc", "GCC=g++", "A=", "TEST=33" ],
         "Shlex replacement test 6!")
@@ -448,6 +481,5 @@ if __name__ == "__main__":
     assertEql(collapse([ "a", "2>&1", '|', 'c' ]), "a 2>&1 | c", "Other separators.")
     assertEql(filterSplitList(shSplit('TEST_MACRO="Testing1234=:= := This **should ** work! "')),
         ['TEST_MACRO="Testing1234=:= := This **should ** work! "'], "Quoting that starts in the middle?")
-    # assertEql(globArgs(['*.*'], ShellState()), ['*.*'], "Test identity globbing")
-    # assertEql(globArgs(['a test', 'of some things', 'that should work'], ShellState()), ['a test', 'of some things', 'that should work'], "Test identity globbing")
+    assertEql(globArgs(['a test', 'of some things', 'that should work'], ShellState()), ['a test', 'of some things', 'that should work'], "Test identity globbing")
     
